package ca.recrown.islandsurvivalcraft.pathfinding;

import java.util.HashSet;
import java.util.PriorityQueue;
import java.util.Queue;

import org.apache.commons.lang.NullArgumentException;

public class DepthFirstSearch {
    private Queue<DFSNode> queue;
    private CoordinateValidatable coordValidatable;
    private DFSNode startNode;
    private DFSNode endNode;
    private HashSet<DFSNode> checkedNodes;
    private int maxNodesSearched = -1;
    private DFSNode foundNode;

    public DepthFirstSearch() {
        checkedNodes = new HashSet<>();
        queue = new PriorityQueue<>();
    }

    public DepthFirstSearch(int maxNodesSearched) {
        this();
        this.maxNodesSearched = maxNodesSearched;
    }

    public DepthFirstSearch(int searchRadius, CoordinateValidatable validatable) {
        this(searchRadius);
        setValidatable(validatable);
    }

    public DepthFirstSearch(CoordinateValidatable validatable) {
        this();
        setValidatable(validatable);
    }

    public void setValidatable(CoordinateValidatable validatable) {
        this.coordValidatable = validatable;
    }

    public void setStartPosition(int x, int y) {
        startNode = new DFSNode(null, x, y, endNode);
    }

    public void setDirectionPosition(int x, int y) {
        endNode = new DFSNode(null, x, y, endNode);
    }

    public DFSNode getFoundNode() {
        return foundNode;
    }

    public void deleteTree() {
        startNode = null;
        endNode = null;
    }

    public boolean buildPathToEndNode() {
        if (coordValidatable == null)
            throw new IllegalStateException("Need to set a validator.");
        if (startNode == null)
            throw new IllegalStateException("Need to set the starting position.");

        return findTarget(new CoordinateTargetValidatable() {

            @Override
            public boolean isCoordinateTarget(int x, int y) {
                return endNode.getX() == x && endNode.getY() == y;
            }
        });
    }

    /**
     * Finds the end node. If an end node is set, it will try going in that
     * direction first. Will not set end node if one wasn't found, therefore,
     * previous end node data is kept.
     * 
     * @param targetValidator
     * @return
     */
    public boolean findTarget(CoordinateTargetValidatable targetValidator) {
        if (coordValidatable == null)
            throw new IllegalStateException("Need to set a validator.");
        if (targetValidator == null)
            throw new NullArgumentException("targetValidator");
        if (startNode == null)
            throw new IllegalStateException("Need to set the starting position.");

        queue.clear();
        checkedNodes.clear();
        DFSNode begin = startNode;
        queue.add(begin);
        while (!queue.isEmpty()) {
            DFSNode n = queue.poll();
            if (maxNodesSearched != -1 && checkedNodes.size() > maxNodesSearched)
                return false;
            if (checkedNodes.add(n) && coordValidatable.validate(n.getX(), n.getY())) {
                n.child[0] = new DFSNode(n, n.getX() + 1, n.getY(), endNode);
                n.child[1] = new DFSNode(n, n.getX() - 1, n.getY(), endNode);
                n.child[2] = new DFSNode(n, n.getX(), n.getY() + 1, endNode);
                n.child[3] = new DFSNode(n, n.getX(), n.getY() - 1, endNode);
                if (targetValidator.isCoordinateTarget(n.getX(), n.getY())) {
                    foundNode = new DFSNode(n.getParent(), n.getX(), n.getY(), null);
                    return true;
                }
                for (int i = 0; i < n.child.length; i++) {
                    DFSNode child = n.child[i];
                    queue.add(child);
                }
            }
        }
        return false;
    }

    public int getSearchedCount() {
        return checkedNodes.size();
    }
}